# -*- coding:utf-8 -*-

"""
WPTools core module
~~~~~~~~~~~~~~~~~~~

Support for accessing Wikimedia foundation APIs.
"""

from time import sleep
import urllib.parse

from wptools.query import WPToolsQuery

from . import utils

import requests
class WPTools(object):
    """
    WPtools (abstract) core class
    """

    REQUEST_DELAY = 0
    REQUEST_LIMIT = 50

    cache = None
    data = None
    flags = None
    params = None

    def __init__(self, *args, **kwargs):
        """
        Abstract initialization for...
        - wptools.page
        - wptools.category
        - wptools.restbase
        - wptools.wikidata
        """
        self.cache = {}
        self.data = {}

        self.flags = {
            'silent': kwargs.get('silent') or False,
            'verbose': kwargs.get('verbose') or False
        }

        self.params = {
            'lang': kwargs.get('lang') or 'en',
        }

        if len(args) > 0 and args[0]:  # first positional arg is title
            self.params.update({'title': args[0]})

        if kwargs.get('skip'):
            self.flags.update({'skip': kwargs.get('skip')})

        if kwargs.get('variant'):
            self.params.update({'variant': kwargs.get('variant')})

        if kwargs.get('wiki'):
            self.params.update({'wiki': kwargs.get('wiki')})

    def _build_showstr(self, seed):
        """
        Returns show() display string for data attribute
        """
        output = ["%s (%s) data" % (seed, self.params['lang'])]

        output.append('{')

        maxwidth = WPToolsQuery.MAXWIDTH

        for item in sorted(self.data):

            if self.data[item] is None:
                continue

            prefix = item
            value = self.data[item]

            if isinstance(value, dict):
                prefix = "%s: <dict(%d)>" % (prefix, len(value))
                value = ', '.join(value.keys())
            elif isinstance(value, int):
                prefix = "%s:" % prefix
                if 'pageid' not in prefix:
                    value = "{:,}".format(value)
            elif isinstance(value, list):
                prefix = "%s: <list(%d)>" % (prefix, len(value))
                value = ', '.join((safestr(x) for x in value if x))
            elif isinstance(value, tuple):
                prefix = "%s: <tuple(%d)>" % (prefix, len(value))
                value = ', '.join((safestr(x) for x in value if x))
            elif utils.is_text(value):
                value = value.strip().replace('\n', '')
                if len(value) > (maxwidth - len(prefix)):
                    prefix = "%s: <str(%d)>" % (prefix, len(value))
                else:
                    prefix = "%s:" % prefix

            output.append("  %s %s" % (prefix, value))

        output.append('}')

        return output

    def _continue_params(self):
        """
        Returns query string fragment continue parameters
        """
        if not self.data.get('continue'):
            return

        params = []
        for item in self.data['continue']:
            params.append("&%s=%s" % (item, urllib.parse.quote_plus(self.data['continue'][item])))

        return ''.join(params)

    def _handle_continuations(self, response, cache_key):
        """
        Select continue params and clear cache or last continue params
        """
        rcontinue = response.get('continue')
        listen = ['blcontinue', 'cmcontinue', 'plcontinue']
        cparams = {}

        if rcontinue:
            for flag in listen:
                if rcontinue.get(flag):
                    cparams[flag] = rcontinue.get(flag)

        if cparams:
            self.data['continue'] = cparams
            del self.cache[cache_key]
        else:  # no more continuations
            if 'continue' in self.data:
                del self.data['continue']

    def _get(self, action, show, proxy, timeout):
        """
        make HTTP request and cache response
        """
        silent = self.flags['silent']

        if action in self.cache:
            if action != 'imageinfo' and action != 'labels':
                utils.stderr("+ %s results in cache" % action, silent)
                return
        else:
            self.cache[action] = {}

        if self.flags.get('skip') and action in self.flags['skip']:
            if not self.flags['silent']:
                utils.stderr("+ skipping %s" % action)
            return

        if 'requests' not in self.data:
            self.data['requests'] = []

        if len(self.data['requests']) >= self.REQUEST_LIMIT:
            raise StopIteration("Hit REQUEST_LIMIT = %d" % self.REQUEST_LIMIT)

        if self.data['requests'] and self.REQUEST_DELAY:
            utils.stderr("REQUEST_DELAY = %d seconds" % self.REQUEST_DELAY)
            sleep(self.REQUEST_DELAY)

        # make the request
        qobj = WPToolsQuery(lang=self.params['lang'],
                            variant=self.params.get('variant'),
                            wiki=self.params.get('wiki'),
                            endpoint=self.params.get('endpoint'))
        qstr = self._query(action, qobj)
        response = requests.get(qstr, qobj.status)
        self.cache[action]['query'] = qstr
        self.cache[action]['response'] = response.json()

        self.data['requests'].append(action)

        self._set_data(action)

        if show and not self.flags.get('silent'):
            self.show()

    def _load_response(self, action):
        """
        returns API reponse from cache or raises ValueError
        """
        _query = self.cache[action]['query'].replace('&format=json', '')
        response = self.cache[action]['response']

        if not response:
            raise ValueError("Empty response: %s" % self.params)

        data = response

        if data.get('warnings'):
            if 'WARNINGS' in self.data:
                self.data['WARNINGS'].update(data['warnings'])
            else:
                self.data['WARNINGS'] = data['warnings']

        if data.get('error'):
            utils.stderr("API error: %s" % data.get('error'))
            raise LookupError(_query)

        if 'query' in action and data.get('query'):
            if data['query'].get('pages'):
                if data['query']['pages'][0].get('missing'):
                    raise LookupError(_query)

        if action == 'parse' and not data.get('parse'):
            raise LookupError(_query)

        if action == 'wikidata':
            handle_wikidata_errors(data, _query)

        return data

    def _query(self, action, qobj):
        """
        Abstract method that returns WPToolsQuery string
        """
        raise NotImplementedError("A subclass must implement this method.")

    def _set_data(self, action):
        """
        Abstract method to capture API response data
        """
        raise NotImplementedError("A subclass must implement this method.")

    def info(self, action=None):
        """
        returns cached request info for given action,
        or list of cached actions
        """
        if action in self.cache:
            return self.cache[action]['info']
        return self.cache.keys() or None

    def query(self, action=None):
        """
        returns cached query string (without &format=json) for given action,
        or list of cached actions
        """
        if action in self.cache:
            return self.cache[action]['query'].replace('&format=json', '')
        return self.cache.keys() or None

    def response(self, action=None):
        """
        returns cached response (as dict) for given action,
        or list of cached actions
        """
        if action in self.cache:
            return self.cache[action]['response']
        return self.cache.keys() or None

    def show(self):
        """
        Pretty-print instance data
        """
        if not self.data:
            return

        if self.data.get('continue'):
            return

        ptitle = self.params.get('title')
        dtitle = self.data.get('title')
        pageid = self.params.get('pageid')

        seed = dtitle or ptitle or pageid
        if utils.is_text(seed):
            seed = seed.replace('_', ' ')

        prettyprint(self._build_showstr(seed))


def handle_wikidata_errors(data, query):
    """
    Raises LookupError if wikidata error found
    """
    entities = data.get('entities')

    if not entities:
        raise LookupError(query)
    elif '-1' in entities:
        raise LookupError(query)
    else:
        item = list(entities.values())[0]
        if 'missing' in item:
            errmsg = "wikidata item %s has been deleted" % item['id']
            raise LookupError(errmsg)


def prettyprint(datastr):
    """
    Print page data strings to stderr
    """
    maxwidth = WPToolsQuery.MAXWIDTH
    rpad = WPToolsQuery.RPAD

    extent = maxwidth - (rpad + 2)
    for line in datastr:
        if len(line) >= maxwidth:
            line = line[:extent] + '...'
        utils.stderr(line)


def safestr(text):
    """
    Safely convert unicode to a string
    """
    if text is None:
        return
    try:
        return str(text)
    except UnicodeEncodeError:
        return str(text.encode('utf-8'))



# -*- coding:utf-8 -*-

"""
WPTools Site module
~~~~~~~~~~~~~~~~~~~

Support for getting Mediawiki site info.
"""

from __future__ import print_function

import random

from . import core


class WPToolsSite(core.WPTools):
    """
    WPToolsSite class
    """

    COMMONS = 'commons.wikimedia.org'

    def __init__(self, *args, **kwargs):
        """
        Returns a WPToolsSite object.

        Optional keyword {params}:
        - [endpoint]: <str> alternative API endpoint (default=/w/api.php)
        - [lang]: <str> Mediawiki language code (default=en)
        - [wiki]: <str> alternative wiki site (default=wikipedia.org)

        Optional keyword {flags}:
        - [silent]: <bool> do not echo page data if True
        - [skip]: <list> skip actions in this list
        - [verbose]: <bool> verbose output to stderr if True
        """
        super(WPToolsSite, self).__init__(*args, **kwargs)

        endpoint = kwargs.get('endpoint')
        if endpoint:
            self.params.update({'endpoint': endpoint})

    def _query(self, action, qobj):
        """
        returns query string
        """
        return qobj.site(action)

    def _set_data(self, action):
        """
        capture Wikidata API response data
        """
        if action == 'siteinfo':
            self._set_siteinfo()
        elif action == 'sitematrix':
            self._set_sitematrix()
        elif action == 'sitevisitors':
            self._set_sitevisitors()

    def _set_siteinfo(self):
        """
        capture API sitematrix data in data attribute
        """
        data = self._load_response('siteinfo').get('query')

        mostviewed = data.get('mostviewed')
        self.data['mostviewed'] = []
        for item in mostviewed[1:]:
            if item['ns'] == 0:
                self.data['mostviewed'].append(item)

        general = data.get('general')

        self.params.update({'title': general.get('sitename')})
        self.params.update({'lang': general.get('lang')})
        self.data['site'] = general.get('wikiid')

        info = {}
        for item in general:
            ginfo = general.get(item)
            if ginfo:
                info[item] = ginfo
        self.data['info'] = info

        siteviews = data.get('siteviews')
        if siteviews:
            values = [x for x in siteviews.values() if x]
            if values:
                self.data['siteviews'] = int(sum(values) / len(values))
            else:
                self.data['siteviews'] = 0

        stats = data.get('statistics')
        for item in stats:
            self.data[item] = stats[item]

    def _set_sitematrix(self):
        """
        capture API sitematrix data in data attribute
        """
        data = self._load_response('sitematrix')

        self.params.update({'title': self.COMMONS})

        matrix = data.get('sitematrix')
        if matrix:
            self.data['sites'] = self._sitelist(matrix)
            self.data['random'] = random.choice(self.data['sites'])

    def _set_sitevisitors(self):
        """
        capture API pageview/visitor data in data attribute
        """
        data = self._load_response('sitevisitors').get('query')

        siteviews = data.get('siteviews')
        if siteviews:
            values = [x for x in siteviews.values() if x]
            if values:
                self.data['visitors'] = int(sum(values) / len(values))
            else:
                self.data['visitors'] = 0

    def _sitelist(self, matrix):
        """
        Returns a list of sites from a SiteMatrix, optionally filtered
        by 'domain' param
        """
        _list = []
        for item in matrix:
            sites = []

            if isinstance(matrix[item], list):
                sites = matrix[item]
            elif isinstance(matrix[item], dict):
                sites = matrix[item]['site']

            for site in sites:
                if len(site.keys()) > 4:  # closed, fishbowl, private
                    continue
                domain = self.params.get('domain')
                if domain:
                    if domain in site['url']:
                        _list.append(site['url'])
                else:
                    _list.append(site['url'])

        return _list

    def get_info(self, wiki=None, show=True, proxy=None, timeout=0):
        """
        GET site info (general, statistics, siteviews, mostviewed) via
        https://www.mediawiki.org/wiki/API:Siteinfo, and
        https://www.mediawiki.org/wiki/Extension:PageViewInfo

        Optional arguments:
        - [wiki]: <str> alternate wiki site (default=en.wikipedia.org)
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - info: <dict> API:Siteinfo
        - mostviewed: <list> mostviewed articles {ns=0, title, count}
        - site: <str> sitename, e.g. 'enwiki'
        - siteviews: <int> sitewide pageview totals over last WEEK
        - visitors: <int> sitewide unique visitor total over last WEEK
        - various counts: activeusers, admins, articles, edits, images
          jobs, pages, queued-massmessages, siteviews, users, visitors
        """
        if wiki:
            self.params.update({'wiki': wiki})

        self._get('siteinfo', show=False, proxy=proxy, timeout=timeout)
        self._get('sitevisitors', show, proxy, timeout)

        return self

    def get_sites(self, domain=None, show=True, proxy=None, timeout=0):
        """
        GET Wikimedia sites via Extension:SiteMatrix
        https://www.mediawiki.org/wiki/Extension:SiteMatrix

        Optional params:
        - [domain]: filter sites on this domain, e.g. 'wiktionary.org'

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - random: randomly selected wiki site
        - sites: <list> of wiki sites (hundreds) from commons SiteMatrix
        """
        if domain:
            self.params.update({'domain': domain})

        self.params.update({'wiki': self.COMMONS})

        self._get('sitematrix', show, proxy, timeout)

        del self.params['wiki']

        return self

    def top(self, wiki=None, limit=25):
        """
        Print list of top viewed articles (ns=0) over the last WEEK
        https://www.mediawiki.org/wiki/Extension:PageViewInfo

        Optional params:
        - [wiki]: <str> alternate wiki site (default=en.wikipedia.org)
        - [limit]: <int> show up to limit articles (max=500)

        See also:
        https://en.wikipedia.org/wiki/Wikipedia_talk:Top_25_Report
        """
        if wiki:
            self.params.update({'wiki': wiki})

        if 'siteinfo' not in self.cache:
            self.get_info(show=False)

        print("%s mostviewed articles:" % (self.data['site']))

        count = 0
        for item in self.data['mostviewed']:
            if item['ns'] == 0:
                count += 1
                print("%d. %s (%s)" % (count, item['title'],
                                       "{:,}".format(item['count'])))
            if count >= limit:
                break



# -*- coding:utf-8 -*-

"""
WPTools Category module
~~~~~~~~~~~~~~~~~~~~~~~

Support for getting Mediawiki category info.
"""

from . import core


class WPToolsCategory(core.WPTools):
    """
    WPToolsCategory class
    """

    def __init__(self, *args, **kwargs):
        """
        Returns a WPToolsCategory object

        Gets a random category without arguments

        Optional positional {params}:
        - [title]: <str> category title

        Optional keyword {params}:
        - [endpoint]: <str> alternative API endpoint (default=/w/api.php)
        - [lang]: <str> Mediawiki language code (default=en)
        - [namespace]: <int> filter members (0=article, 14=category)
        - [pageid]: <int> category pageid
        - [variant]: <str> Mediawiki language variant
        - [wiki]: <str> alternative wiki site (default=wikipedia.org)

        Optional keyword {flags}:
        - [silent]: <bool> do not echo page data if True
        - [skip]: <list> skip actions in this list
        - [verbose]: <bool> verbose output to stderr if True

        See also:
        https://www.mediawiki.org/wiki/Manual:Namespace
        """
        super(WPToolsCategory, self).__init__(**kwargs)

        endpoint = kwargs.get('endpoint')
        pageid = kwargs.get('pageid')
        namespace = kwargs.get('namespace')

        title = None
        if len(args) > 0:
            title = args[0]
            self.params.update({'title': title})

        if endpoint:
            self.params.update({'endpoint': endpoint})

        if pageid:
            try:
                self.params.update({'pageid': int(pageid)})
            except ValueError:
                raise ValueError("invalid pageid")
            if title:
                raise ValueError("cannot use both title AND pageid")

        if namespace or namespace == 0:
            try:
                self.params.update({'namespace': int(namespace)})
            except ValueError:
                raise ValueError("invalid namespace")

        if not pageid and not title:
            self.get_random()

    def _add_members(self, catmembers):
        """
        Adds category members and subcategories to data
        """
        members = [x for x in catmembers if x['ns'] == 0]
        subcats = [x for x in catmembers if x['ns'] == 14]

        if 'members' in self.data:
            self.data['members'].extend(members)
        else:
            self.data.update({'members': members})

        if subcats:
            if 'subcategories' in self.data:
                self.data['subcategories'].extend(subcats)
            else:
                self.data.update({'subcategories': subcats})

    def _query(self, action, qobj):
        """
        Form query to enumerate category
        """
        title = self.params.get('title')
        pageid = self.params.get('pageid')

        if action == 'random':
            return qobj.random(namespace=14)
        elif action == 'category':
            return qobj.category(title, pageid, self._continue_params())

    def _set_data(self, action):
        """
        Set category member data from API response
        """
        data = self._load_response(action)

        self._handle_continuations(data, 'category')

        if action == 'category':
            members = data.get('query').get('categorymembers')
            if members:
                self._add_members(members)

        if action == 'random':
            rand = data['query']['random'][0]
            data = {'pageid': rand.get('id'),
                    'title': rand.get('title')}
            self.data.update(data)
            self.params.update(data)

    def get_members(self, show=True, proxy=None, timeout=0):
        """
        GET Mediawiki:API (action=query) category members
        https://www.mediawiki.org/wiki/API:Categorymembers

        Required {params}: title OR pageid
        - title: <str> article title
        - pageid: <int> Wikipedia database ID

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - members: <list> category members [{ns, pageid, title}]
        """
        title = self.params.get('title')
        pageid = self.params.get('pageid')

        if not title and not pageid:
            raise LookupError("needs category title or pageid")

        self._get('category', show, proxy, timeout)

        while self.data.get('continue'):
            self._get('category', show, proxy, timeout)

        return self

    def get_random(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki:API (action=query) for random category
        https://www.mediawiki.org/wiki/API:Random

        Required {params}: None

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - pageid: <int> Wikipedia database ID
        - title: <str> article title
        """
        self._get('random', show, proxy, timeout)

        # flush cache to allow repeated random requests
        del self.cache['random']

        return self



# -*- coding:utf-8 -*-

query = 'https://en.wikipedia.org/w/api.php?action=query&exintro&formatversion=2&inprop=url|watchers&list=random&pithumbsize=240&pllimit=500&ppprop=disambiguation|wikibase_item&prop=extracts|info|links|pageimages|pageprops|pageterms&redirects&rnlimit=1&rnnamespace=0&titles=Douglas%20Adams%20%28disambiguation%29'

response = r"""
{
    "batchcomplete": true,
    "continue": {
        "rncontinue": "0.841264018333|0.841264104557|48635964|0",
        "continue": "-||extracts|info|links|pageimages|pageprops|pageterms"
    },
    "query": {
        "pages": [
            {
                "pageid": 1167171,
                "ns": 0,
                "title": "Douglas Adams (disambiguation)",
                "extract": "<p><b>Douglas Adams</b> was an English writer and dramatist.</p>\n<p><b>Douglas</b> or <b>Doug Adams</b> may also refer to:</p>\n<ul><li>Douglas Adams (engineer), American engineer</li>\n<li>Douglas Q. Adams, Indo-Europeanist professor of English</li>\n<li>Doug Adams (American football) (1949–1997), American football player</li>\n<li>Doug Adams (baseball) (born 1943), MLB player</li>\n<li>Doug Adams (music journalist), music journalist and author</li>\n<li>Doug Adams (television producer), American television producer</li>\n<li>Douglas Adams, CEO of Adams Cable</li>\n<li>Douglas Adams (cricketer) (1876–1931), American cricketer</li>\n</ul>",
                "contentmodel": "wikitext",
                "pagelanguage": "en",
                "pagelanguagehtmlcode": "en",
                "pagelanguagedir": "ltr",
                "touched": "2017-10-18T01:46:38Z",
                "lastrevid": 805854377,
                "length": 671,
                "fullurl": "https://en.wikipedia.org/wiki/Douglas_Adams_(disambiguation)",
                "editurl": "https://en.wikipedia.org/w/index.php?title=Douglas_Adams_(disambiguation)&action=edit",
                "canonicalurl": "https://en.wikipedia.org/wiki/Douglas_Adams_(disambiguation)",
                "links": [
                    {
                        "ns": 0,
                        "title": "Adams Cable"
                    },
                    {
                        "ns": 0,
                        "title": "Doug Adam"
                    },
                    {
                        "ns": 0,
                        "title": "Doug Adams (American football)"
                    },
                    {
                        "ns": 0,
                        "title": "Doug Adams (baseball)"
                    },
                    {
                        "ns": 0,
                        "title": "Doug Adams (music journalist)"
                    },
                    {
                        "ns": 0,
                        "title": "Doug Adams (television producer)"
                    },
                    {
                        "ns": 0,
                        "title": "Douglas Adams"
                    },
                    {
                        "ns": 0,
                        "title": "Douglas Adams (cricketer)"
                    },
                    {
                        "ns": 0,
                        "title": "Douglas Adams (engineer)"
                    },
                    {
                        "ns": 0,
                        "title": "Douglas Q. Adams"
                    },
                    {
                        "ns": 1,
                        "title": "Talk:Douglas Adams (disambiguation)"
                    },
                    {
                        "ns": 12,
                        "title": "Help:Disambiguation"
                    }
                ],
                "pageprops": {
                    "disambiguation": "",
                    "wikibase_item": "Q5301194"
                },
                "terms": {
                    "description": [
                        "Wikipedia disambiguation page"
                    ],
                    "label": [
                        "Douglas Adams"
                    ]
                }
            }
        ],
        "random": [
            {
                "id": 50798021,
                "ns": 0,
                "title": "Morchella arbutiphila"
            }
        ]
    }
}
"""

cache = {'info': {'status': 200}, 'query': query, 'response': response}



# -*- coding:utf-8 -*-

query = 'https://www.wikidata.org/w/api.php?action=wbgetentities&formatversion=2&languages=en&props=labels&redirects=yes&ids=P4193|P119|Q14623673|P691|P856|P18|P1559|P3373|P2048|P509|P569|P345|P1006|P349|P1005|P1003|P2611|P2191|Q12152|P109|P108|Q42406|P106|Q3739104|P269|P103|P25|P27|P26|P21|P20|P22|P866|Q6581097|Q691283|P2188|P646|Q7979|Q28389|Q4853732|P800|P373|P648|Q128758|Q14623681|P2626|P1233|P1442|P570|Q1042294'

response = r"""{
    "entities": {
        "P4193": {
            "type": "property",
            "datatype": "external-id",
            "id": "P4193",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Familypedia person ID"
                }
            }
        },
        "P119": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P119",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "place of interment"
                }
            }
        },
        "Q14623673": {
            "type": "item",
            "id": "Q14623673",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Susan Adams"
                }
            }
        },
        "P691": {
            "type": "property",
            "datatype": "external-id",
            "id": "P691",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "NKCR AUT ID"
                }
            }
        },
        "P856": {
            "type": "property",
            "datatype": "url",
            "id": "P856",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "official website"
                }
            }
        },
        "P18": {
            "type": "property",
            "datatype": "commonsMedia",
            "id": "P18",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "image"
                }
            }
        },
        "P1559": {
            "type": "property",
            "datatype": "monolingualtext",
            "id": "P1559",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "name in native language"
                }
            }
        },
        "P3373": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P3373",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "sibling"
                }
            }
        },
        "P2048": {
            "type": "property",
            "datatype": "quantity",
            "id": "P2048",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "height"
                }
            }
        },
        "P509": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P509",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "cause of death"
                }
            }
        },
        "P569": {
            "type": "property",
            "datatype": "time",
            "id": "P569",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "date of birth"
                }
            }
        },
        "P345": {
            "type": "property",
            "datatype": "external-id",
            "id": "P345",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "IMDb ID"
                }
            }
        },
        "P1006": {
            "type": "property",
            "datatype": "external-id",
            "id": "P1006",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "National Thesaurus for Author Names ID"
                }
            }
        },
        "P349": {
            "type": "property",
            "datatype": "external-id",
            "id": "P349",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "NDL Auth ID"
                }
            }
        },
        "P1005": {
            "type": "property",
            "datatype": "external-id",
            "id": "P1005",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "PTBNP ID"
                }
            }
        },
        "P1003": {
            "type": "property",
            "datatype": "external-id",
            "id": "P1003",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "NLR (Romania) ID"
                }
            }
        },
        "P2611": {
            "type": "property",
            "datatype": "external-id",
            "id": "P2611",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "TED speaker ID"
                }
            }
        },
        "P2191": {
            "type": "property",
            "datatype": "external-id",
            "id": "P2191",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "NILF author id"
                }
            }
        },
        "Q12152": {
            "type": "item",
            "id": "Q12152",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "myocardial infarction"
                }
            }
        },
        "P109": {
            "type": "property",
            "datatype": "commonsMedia",
            "id": "P109",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "signature"
                }
            }
        },
        "P108": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P108",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "employer"
                }
            }
        },
        "Q42406": {
            "type": "item",
            "id": "Q42406",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "English people"
                }
            }
        },
        "P106": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P106",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "occupation"
                }
            }
        },
        "Q3739104": {
            "type": "item",
            "id": "Q3739104",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "natural causes"
                }
            }
        },
        "P269": {
            "type": "property",
            "datatype": "external-id",
            "id": "P269",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "SUDOC authorities"
                }
            }
        },
        "P103": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P103",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "native language"
                }
            }
        },
        "P25": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P25",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "mother"
                }
            }
        },
        "P27": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P27",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "country of citizenship"
                }
            }
        },
        "P26": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P26",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "spouse"
                }
            }
        },
        "P21": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P21",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "sex or gender"
                }
            }
        },
        "P20": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P20",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "place of death"
                }
            }
        },
        "P22": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P22",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "father"
                }
            }
        },
        "P866": {
            "type": "property",
            "datatype": "external-id",
            "id": "P866",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Perlentaucher ID"
                }
            }
        },
        "Q6581097": {
            "type": "item",
            "id": "Q6581097",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "male"
                }
            }
        },
        "Q691283": {
            "type": "item",
            "id": "Q691283",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "St John's College"
                }
            }
        },
        "P2188": {
            "type": "property",
            "datatype": "external-id",
            "id": "P2188",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "BiblioNet author ID"
                }
            }
        },
        "P646": {
            "type": "property",
            "datatype": "external-id",
            "id": "P646",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Freebase ID"
                }
            }
        },
        "Q7979": {
            "type": "item",
            "id": "Q7979",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "British English"
                }
            }
        },
        "Q28389": {
            "type": "item",
            "id": "Q28389",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "screenwriter"
                }
            }
        },
        "Q4853732": {
            "type": "item",
            "id": "Q4853732",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "children's writer"
                }
            }
        },
        "P800": {
            "type": "property",
            "datatype": "wikibase-item",
            "id": "P800",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "notable work"
                }
            }
        },
        "P373": {
            "type": "property",
            "datatype": "string",
            "id": "P373",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Commons category"
                }
            }
        },
        "P648": {
            "type": "property",
            "datatype": "external-id",
            "id": "P648",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Open Library ID"
                }
            }
        },
        "Q128758": {
            "type": "item",
            "id": "Q128758",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "satire"
                }
            }
        },
        "Q14623681": {
            "type": "item",
            "id": "Q14623681",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "Jane Belson"
                }
            }
        },
        "P2626": {
            "type": "property",
            "datatype": "external-id",
            "id": "P2626",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "DNF person ID"
                }
            }
        },
        "P1233": {
            "type": "property",
            "datatype": "external-id",
            "id": "P1233",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "ISFDB author ID"
                }
            }
        },
        "P1442": {
            "type": "property",
            "datatype": "commonsMedia",
            "id": "P1442",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "image of grave"
                }
            }
        },
        "P570": {
            "type": "property",
            "datatype": "time",
            "id": "P570",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "date of death"
                }
            }
        },
        "Q1042294": {
            "type": "item",
            "id": "Q1042294",
            "labels": {
                "en": {
                    "language": "en",
                    "value": "So Long, and Thanks for All the Fish"
                }
            }
        }
    },
    "success": 1
}"""

cache = {'info': {'status': 200}, 'query': query, 'response': response}



# -*- coding:utf-8 -*-

"""
WPTools Query module
~~~~~~~~~~~~~~~~~~~~

Support for forming WMF API query strings.

* Mediawiki: https://www.mediawiki.org/wiki/API:Main_page
* Wikidata: https://www.wikidata.org/wiki/Wikidata:Data_access
* RESTBase: https://www.mediawiki.org/wiki/RESTBase

See also:

* WMF: https://wikimediafoundation.org/wiki/Our_projects
"""

try:  # python2
    from urllib import quote, unquote
except ImportError:  # python3
    from urllib.parse import quote, unquote

from string import Template

import random


class WPToolsQuery(object):
    """
    WPToolsQuery class
    """

    DEFAULT_ENDPOINT = '/w/api.php'
    MAXWIDTH = 72
    RPAD = 4

    IMAGEINFO = Template((
        "${WIKI}${ENDPOINT}?action=query"
        "&format=json"
        "&formatversion=2"
        "&iiprop=size|url|timestamp|extmetadata"
        "&prop=imageinfo"
        "&titles=${FILES}"))

    LIST = Template((
        "${WIKI}${ENDPOINT}?action=query"
        "&format=json"
        "&formatversion=2"
        "&list=${LIST}"))

    PARSE = Template((
        "${WIKI}${ENDPOINT}?action=parse"
        "&format=json"
        "&formatversion=2"
        "&contentmodel=text"
        "&disableeditsection="
        "&disablelimitreport="
        "&disabletoc="
        "&prop=text|iwlinks|parsetree|wikitext|displaytitle|properties"
        "&redirects"
        "&page=${PAGE}"))

    QUERY = Template((
        "${WIKI}${ENDPOINT}?action=query"
        "&exintro"
        "&format=json"
        "&formatversion=2"
        "&inprop=url|watchers"
        "&list=random"
        "&pithumbsize=240"
        "&pllimit=500"
        "&ppprop=disambiguation|wikibase_item"
        "&prop=extracts|info|links|pageassessments|pageimages|pageprops"
        "|pageterms|redirects"
        "&redirects"
        "&rdlimit=500"
        "&rnlimit=1"
        "&rnnamespace=0"
        "&titles=${TITLES}"))

    QUERYMORE = Template((
        "${WIKI}${ENDPOINT}?action=query"
        "&bllimit=500"
        "&bltitle=${TITLES}"
        "&cllimit=500"
        "&clshow=!hidden"
        "&format=json"
        "&formatversion=2"
        "&imlimit=500"
        "&list=backlinks"
        "&lllimit=500"
        "&pclimit=500"
        "&prop=categories|contributors|images|langlinks|pageviews"
        "&redirects"
        "&titles=${TITLES}"))

    WIKIDATA = Template((
        "${WIKI}${ENDPOINT}?action=wbgetentities"
        "&format=json"
        "&formatversion=2"
        "&languages=${LANG}"
        "&props=${PROPS}"
        "&redirects=yes"))

    endpoint = None
    lang = None
    status = None
    variant = None
    wiki = None

    def __init__(self, lang='en', variant=None, wiki=None, endpoint=None):
        """
        Returns a WPToolsQuery object

        Arguments:
        - [lang=en]: <str> Mediawiki language code
        - [variant=None]: <str> language variant
        - [wiki=None]: <str> alternative wiki site
        - [endpoint=None]: <str> alternative API endoint
        """
        self.lang = lang
        self.variant = variant

        self.wiki = wiki or "%s.wikipedia.org" % self.lang
        self.domain = domain_name(self.wiki)
        self.endpoint = endpoint or self.DEFAULT_ENDPOINT
        self.uri = self.wiki_uri(self.wiki)

    def category(self, title, pageid=None, cparams=None, namespace=None):
        """
        Returns category query string
        """
        query = self.LIST.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            LIST='categorymembers')
        status = pageid or title

        query += "&cmlimit=500"

        if namespace is not None:
            query += "&cmnamespace=%d" % namespace

        if title and pageid:
            title = None

        if title:
            query += "&cmtitle=" + safequote(title)

        if pageid:
            query += "&cmpageid=%d" % pageid

        if cparams:
            query += cparams
            status += ' (%s)' % cparams

        self.set_status('categorymembers', status)

        return query

    def labels(self, qids):
        """
        Returns Wikidata labels query string
        """
        if len(qids) > 50:
            raise ValueError("The limit is 50.")

        self.domain = 'www.wikidata.org'
        self.uri = self.wiki_uri(self.domain)

        query = self.WIKIDATA.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            LANG=self.variant or self.lang,
            PROPS='labels')

        qids = '|'.join(qids)
        query += "&ids=%s" % qids

        self.set_status('labels', qids)

        return query

    def imageinfo(self, files):
        """
        Returns imageinfo query string
        """
        files = '|'.join([safequote(x) for x in files])

        self.set_status('imageinfo', files)

        return self.IMAGEINFO.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            FILES=files)

    def parse(self, title, pageid=None):
        """
        Returns Mediawiki action=parse query string
        """
        qry = self.PARSE.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            PAGE=safequote(title) or pageid)

        if pageid and not title:
            qry = qry.replace('&page=', '&pageid=').replace('&redirects', '')

        if self.variant:
            qry += '&variant=' + self.variant

        self.set_status('parse', pageid or title)

        return qry

    def query(self, titles, pageids=None, cparams=None):
        """
        Returns MediaWiki action=query query string
        """
        query = self.QUERY.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            TITLES=safequote(titles) or pageids)
        status = titles or pageids

        if pageids and not titles:
            query = query.replace('&titles=', '&pageids=')

        if cparams:
            query += cparams
            status += " (%s)" % cparams

        if self.variant:
            query += '&variant=' + self.variant

        self.set_status('query', status)

        return query

    def querymore(self, titles, pageids=None, cparams=None):
        """
        Returns MediaWiki action=query query string (for MORE)
        A much more expensive query for popular pages
        """
        query = self.QUERYMORE.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            TITLES=safequote(titles) or pageids)

        status = "%s" % (pageids or titles)

        if pageids and not titles:
            query = query.replace('&titles=', '&pageids=')

        if cparams:
            query += cparams
            status += " (%s)" % cparams

        if self.variant:
            query += '&variant=' + self.variant

        self.set_status('querymore', status)

        return query

    def random(self, namespace=0):
        """
        Returns query string for random page
        """
        query = self.LIST.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            LIST='random')
        query += "&rnlimit=1&rnnamespace=%d" % namespace

        emoji = [
            u'\U0001f32f',  # burrito or wrap
            u'\U0001f355',  # slice of pizza
            u'\U0001f35c',  # steaming bowl of ramen
            u'\U0001f363',  # sushi
            u'\U0001f369',  # doughnut
            u'\U0001f36a',  # cookie
            u'\U0001f36d',  # lollipop
            u'\U0001f370',  # strawberry shortcake
        ]

        action = 'random'
        if namespace:
            action = 'random:%d' % namespace

        self.set_status(action, random.choice(emoji))

        return query

    def restbase(self, endpoint, title):
        """
        Returns RESTBase query string
        """
        if not endpoint:
            raise ValueError("invalid endpoint: %s" % endpoint)

        route = endpoint
        if title and endpoint != '/page/':
            route = endpoint + safequote_restbase(title)

        self.set_status('restbase', route)

        return "%s/api/rest_v1/%s" % (self.uri, route[1:])

    def set_status(self, action, target):
        """
        Sets query status with format: "{domain} ({action}) {target}"
        """
        try:
            target = unquote(target)
        except (AttributeError, TypeError):
            pass

        status = "%s (%s) %s" % (self.domain, action, target)
        status = status.strip().replace('\n', '')

        if len(status) >= self.MAXWIDTH:
            tail = '...'
            extent = self.MAXWIDTH - (len(tail) + self.RPAD)
            self.status = status[:extent] + tail
        else:
            self.status = status

    def site(self, action):
        """
        Returns site query
        """
        query = None
        viewdays = 7
        hostpath = self.uri + self.endpoint

        if action == 'siteinfo':
            query = hostpath + (
                '?action=query'
                '&meta=siteinfo|siteviews'
                '&siprop=general|statistics'
                '&list=mostviewed&pvimlimit=max')
            query += '&pvisdays=%d' % viewdays  # meta=siteviews
            self.set_status('query', 'siteinfo|siteviews|mostviewed')
        elif action == 'sitematrix':
            query = hostpath + '?action=sitematrix'
            self.set_status('sitematrix', 'all')
        elif action == 'sitevisitors':
            query = hostpath + (
                '?action=query'
                '&meta=siteviews&pvismetric=uniques')
            query += '&pvisdays=%d' % viewdays  # meta=siteviews
            self.set_status('query', 'siteviews:uniques')

        if not query:
            raise ValueError("Could not form query")

        query += '&format=json&formatversion=2'

        return query

    def wiki_uri(self, wiki):
        """
        Returns scheme://domain from wiki name
        """
        if wiki.startswith('http'):
            return wiki
        return "https://" + self.domain

    def wikidata(self, title, wikibase=None):
        """
        Returns Wikidata query string
        """
        self.domain = 'www.wikidata.org'
        self.uri = self.wiki_uri(self.domain)

        query = self.WIKIDATA.substitute(
            WIKI=self.uri,
            ENDPOINT=self.endpoint,
            LANG=self.variant or self.lang,
            PROPS="aliases|info|claims|descriptions|labels|sitelinks")

        if wikibase:
            query += "&ids=%s" % wikibase
        elif title:
            title = safequote(title)
            query += "&sites=%swiki" % self.lang
            query += "&titles=%s" % title

        self.set_status('wikidata', wikibase or title)

        return query


def domain_name(wiki):
    """
    Returns domain name from wiki name
    """
    if '//' in wiki:
        wiki = wiki.split('//')[1]
    return wiki.split('/')[0]


def safequote(string):
    """
    Try to UTF-8 encode and percent-quote string
    """
    if string is None:
        return
    try:
        return quote(string.encode('utf-8'))
    except UnicodeDecodeError:
        return quote(string)


def safequote_restbase(title):
    """
    Safequote restbase title possibly having slash in title
    """
    try:
        return quote(title.encode('utf-8'), safe='')
    except UnicodeDecodeError:
        return quote(title, safe='')



# -*- coding: utf-8 -*-

"""
Wikipedia tools (for Humans)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Python and command-line MediaWiki access for Humans

- get page extracts, image, Infobox data, Wikidata, and more
- get a random page, category, or site
- get page statistics
- get category members
- get site info and stats
- get data in any language
"""

__author__ = "siznax"
__contact__ = "https://github.com/siznax/wptools"
__license__ = "MIT"
__title__ = "wptools"
__version__ = "0.4.17"

from . import core
from . import query
#from . import request
from . import site
from . import utils

from .category import WPToolsCategory as category
from .page import WPToolsPage as page
from .restbase import WPToolsRESTBase as restbase
from .site import WPToolsSite as site
from .wikidata import WPToolsWikidata as wikidata



# -*- coding:utf-8 -*-

"""
WPTools Page module
~~~~~~~~~~~~~~~~~~~

Support for getting Wikimedia page info.

Access to Wikimedia APIs:

- Mediawiki: https://www.mediawiki.org/wiki/API:Main_page
- RESTBase: https://www.mediawiki.org/wiki/RESTBase
- Wikidata: https://www.wikidata.org/wiki/Wikidata:Data_access

See also:

- https://www.mediawiki.org/wiki/Manual:Page_table
"""

import html2text

from . import core
from . import utils

from .restbase import WPToolsRESTBase
from .wikidata import WPToolsWikidata


class WPToolsPage(WPToolsRESTBase,
                  WPToolsWikidata,
                  core.WPTools):
    """
    WPtools Page class, derived from wptools.core
    """

    def __init__(self, *args, **kwargs):
        """
        Returns a WPToolsPage object

        Gets a random title without arguments

        Optional positional {params}:
        - [title]: <str> Mediawiki page title, file, category, etc.

        Optional keyword {params}:
        - [boxterm]: <str> Infobox title name or substring
        - [endpoint]: <str> alternative API endpoint (default=/w/api.php)
        - [lang]: <str> Mediawiki language code (default=en)
        - [pageid]: <int> Mediawiki pageid
        - [variant]: <str> Mediawiki language variant
        - [wiki]: <str> alternative wiki site (default=wikipedia.org)
        - [wikibase]: <str> Wikidata database ID (e.g. 'Q1')

        Optional keyword {flags}:
        - [silent]: <bool> do not echo page data if True
        - [skip]: <list> skip actions in this list
        - [verbose]: <bool> verbose output to stderr if True
        """
        super(WPToolsPage, self).__init__(*args, **kwargs)

        title = self.params.get('title')

        boxterm = kwargs.get('boxterm')
        if boxterm:
            self.params.update({'boxterm': boxterm})

        endpoint = kwargs.get('endpoint')
        if endpoint:
            self.params.update({'endpoint': endpoint})

        pageid = kwargs.get('pageid')
        if pageid:
            self.params.update({'pageid': pageid})

        wikibase = kwargs.get('wikibase')
        if wikibase:
            self.params.update({'wikibase': wikibase})

        if not title and not pageid and not wikibase:
            self.get_random()
        else:
            self.show()

    def __insert_image_info(self, title, _from, info):
        """
        Insert API image INFO into matching image dict

        We make one imageinfo request containing only unique image
        filenames. We reduce duplication by asking for image data per
        file, instead of per "kind" or source (Wikipedia, Wikidata,
        etc.), because some sources reference the same image file. We
        match API imageinfo response data to existing image filenames
        by API title or normalized "from" title. So, some imageinfo
        data will be applied to more than one image "kind" (source) if
        they share the same filename.
        """
        for img in self.data['image']:
            if 'url' not in img:
                if title == img['file']:  # matching title/file
                    img.update(info)
                elif _from == img['file']:  # matching from/file
                    img.update(info)

    def __pull_image_info(self, title, imageinfo, normalized):
        """
        Pull image INFO from API response and insert
        """
        for info in imageinfo:
            info.update({'title': title})

            # get API normalized "from" filename for matching
            _from = None
            for norm in normalized:
                if title == norm['to']:
                    _from = norm['from']

            # let's put all "metadata" in one member
            info['metadata'] = {}
            extmetadata = info.get('extmetadata')
            if extmetadata:
                info['metadata'].update(extmetadata)
                del info['extmetadata']

            self.__insert_image_info(title, _from, info)

    def _extend_data(self, datapoint, new_data):
        """
        extend or assign new data to datapoint
        """
        if new_data:
            try:
                self.data[datapoint].extend(new_data)
            except KeyError:
                self.data[datapoint] = new_data

    def _missing_imageinfo(self):
        """
        returns list of image filenames that are missing info
        """
        if 'image' not in self.data:
            return
        missing = []
        for img in self.data['image']:
            if 'url' not in img:
                missing.append(img['file'])
        return list(set(missing))

    def _normalize_images(self):
        """
        normalizes image filenames by prepending 'File:' if needed
        """
        if 'image' not in self.data:
            return
        for img in self.data['image']:
            fname = img['file'].replace('_', ' ')
            fstart = fname.startswith('File:')
            istart = fname.startswith('Image:')
            if not fstart and not istart:
                fname = 'File:' + fname
                img['orig'] = img['file']
                img['file'] = fname

    def _query(self, action, qobj):
        """
        returns WPToolsQuery string
        """
        title = self.params.get('title')
        pageid = self.params.get('pageid')
        wikibase = self.params.get('wikibase')

        qstr = None

        if action == 'random':
            qstr = qobj.random()
        elif action == 'query':
            qstr = qobj.query(title, pageid, self._continue_params())
        elif action == 'querymore':
            qstr = qobj.querymore(title, pageid, self._continue_params())
        elif action == 'parse':
            qstr = qobj.parse(title, pageid)
        elif action == 'imageinfo':
            qstr = qobj.imageinfo(self._missing_imageinfo())
        elif action == 'labels':
            qstr = qobj.labels(self._pop_entities())
        elif action == 'wikidata':
            qstr = qobj.wikidata(title, wikibase)
        elif action == 'restbase':
            qstr = qobj.restbase(self.params.get('rest_endpoint'), title)

        if qstr is None:
            raise ValueError("Unknown action: %s" % action)

        return qstr

    def _set_data(self, action):
        """
        marshals response data into page data
        """
        if 'query' in action:
            self._set_query_data(action)
        elif action == 'imageinfo':
            self._set_imageinfo_data()
        elif action == 'parse':
            self._set_parse_data()
        elif action == 'random':
            self._set_random_data()
        elif action == 'labels':
            self._set_labels()
        elif action == 'wikidata':
            self._set_wikidata()
            self.get_labels()
        elif action == 'restbase':
            self._set_restbase_data()

        self._update_imageinfo()
        self._update_params()

    def _set_imageinfo_data(self):
        """
        set image attributes from MediaWiki API:Imageinfo response
        """
        data = self._load_response('imageinfo')
        pages = data['query'].get('pages')

        normalized = []
        if 'normalized' in data['query']:
            normalized = data['query']['normalized']

        for page in pages:
            title = page.get('title')
            imageinfo = page.get('imageinfo')
            if imageinfo:
                self.__pull_image_info(title, imageinfo, normalized)

        # Mark missing imageinfo to prevent duplicate requests
        for img in self.data['image']:
            if 'url' not in img:
                img['url'] = 'MISSED'

    def _set_parse_data(self):
        """
        set attributes derived from MediaWiki (action=parse)
        """
        pdata = self._load_response('parse')['parse']

        self.data['iwlinks'] = utils.get_links(pdata.get('iwlinks'))
        self.data['pageid'] = pdata.get('pageid')
        self.data['wikitext'] = pdata.get('wikitext')

        parsetree = pdata.get('parsetree')
        self.data['parsetree'] = parsetree

        boxterm = self.params.get('boxterm')
        if boxterm:
            infobox = utils.get_infobox(parsetree, boxterm)
        else:
            infobox = utils.get_infobox(parsetree)
        self.data['infobox'] = infobox

        title = pdata.get('title')
        if title:
            self.data['title'] = title
            if not self.params.get('title'):
                self.params['title'] = title

        wikibase = pdata.get('properties').get('wikibase_item')
        if wikibase:
            self.data['wikibase'] = wikibase
            self.data['wikidata_url'] = utils.wikidata_url(wikibase)

        if self.data['infobox']:
            self._set_parse_image(self.data['infobox'])

    def _set_parse_image(self, infobox):
        """
        set image data from action=parse response
        """
        image = infobox.get('image')
        cover = infobox.get('Cover') or infobox.get('cover')

        if image or cover:
            if 'image' not in self.data:
                self.data['image'] = []

        if image and utils.isfilename(image):
            self.data['image'].append({'kind': 'parse-image', 'file': image})

        if cover and utils.isfilename(cover):
            self.data['image'].append({'kind': 'parse-cover', 'file': cover})

    def _set_query_data(self, action='query'):
        """
        set attributes derived from MediaWiki (action=query)
        """
        data = self._load_response(action)
        page = data['query']['pages'][0]

        self._handle_continuations(data, action)

        if action == 'query':
            self.data['random'] = data['query']['random'][0]["title"]

        self._extend_data('backlinks', data['query'].get('backlinks'))

        self.data['redirected'] = data['query'].get('redirects')

        self._set_query_data_fast_1(page)  # avoid pylint too-many-branches
        self._set_query_data_fast_2(page)
        self._set_query_data_slow(page)

    def _set_query_data_fast_1(self, page):
        """
        set less expensive action=query response data PART 1
        """
        self.data['pageid'] = page.get('pageid')

        assessments = page.get('pageassessments')
        if assessments:
            self.data['assessments'] = assessments

        extract = page.get('extract')
        if extract:
            self.data['extract'] = extract
            extext = html2text.html2text(extract)
            if extext:
                self.data['extext'] = extext.strip()

        fullurl = page.get('fullurl')
        if fullurl:
            self.data['url'] = fullurl
            self.data['url_raw'] = fullurl + '?action=raw'

        length = page.get('length')
        if length:
            self.data['length'] = length

        self._extend_data('links', utils.get_links(page.get('links')))

        self._update_data('modified', 'page', page.get('touched'))

        pageprops = page.get('pageprops')
        if pageprops:
            wikibase = pageprops.get('wikibase_item')
            if wikibase:
                self.data['wikibase'] = wikibase
                self.data['wikidata_url'] = utils.wikidata_url(wikibase)

            if 'disambiguation' in pageprops:
                self.data['disambiguation'] = len(self.data['links'])

    def _set_query_data_fast_2(self, page):
        """
        set less expensive action=query response data PART 2
        """
        self.data['pageid'] = page.get('pageid')

        redirects = page.get('redirects')
        if redirects:
            self.data['redirects'] = redirects

        terms = page.get('terms')
        if terms:
            if terms.get('alias'):
                self.data['aliases'] = terms['alias']

            if terms.get('description'):
                self.data['description'] = next(iter(terms['description']),
                                                None)
            if terms.get('label'):
                self.data['label'] = next(iter(terms['label']), None)

        title = page.get('title')
        self.data['title'] = title
        if not self.params.get('title'):
            self.params['title'] = title

        watchers = page.get('watchers')
        if watchers:
            self.data['watchers'] = watchers

        self._set_query_image(page)

    def _set_query_data_slow(self, page):
        """
        set more expensive action=query response data
        """
        categories = page.get('categories')
        if categories:
            self.data['categories'] = [x['title'] for x in categories]

        if page.get('contributors'):
            contributors = page.get('contributors') or 0
            anoncontributors = page.get('anoncontributors') or 0
            if isinstance(contributors, list):
                contributors = len(contributors)
            self.data['contributors'] = contributors + anoncontributors

        files = page.get('images')  # really, these are FILES
        if files:
            self.data['files'] = [x['title'] for x in files]

        languages = page.get('langlinks')
        if languages:
            self.data['languages'] = languages

        pageviews = page.get('pageviews')
        if pageviews:
            values = [x for x in pageviews.values() if x]
            if values:
                self.data['views'] = int(sum(values) / len(values))
            else:
                self.data['views'] = 0

    def _set_query_image(self, page):
        """
        set image data from action=query response
        """
        pageimage = page.get('pageimage')
        thumbnail = page.get('thumbnail')

        if pageimage or thumbnail:
            if 'image' not in self.data:
                self.data['image'] = []

        if pageimage:
            self.data['image'].append({
                'kind': 'query-pageimage',
                'file': pageimage})

        if thumbnail:
            qthumb = {'kind': 'query-thumbnail'}
            qthumb.update(thumbnail)
            qthumb['url'] = thumbnail.get('source')
            del qthumb['source']
            qthumb['file'] = qthumb['url'].split('/')[-2]
            self.data['image'].append(qthumb)

    def _set_random_data(self):
        """
        sets page data from random request
        """
        rdata = self._load_response('random')
        rdata = rdata['query']['random'][0]

        pageid = rdata.get('id')
        title = rdata.get('title')

        self.data.update({'pageid': pageid,
                          'title': title})

    def _update_data(self, datapoint, key, new_data):
        """
        update or assign new data to datapoint
        """
        if new_data:
            try:
                self.data[datapoint].update({key: new_data})
            except KeyError:
                self.data[datapoint] = {key: new_data}

    def _update_imageinfo(self):
        """
        calls get_imageinfo() if data image missing info
        """
        missing = self._missing_imageinfo()
        deferred = self.flags.get('defer_imageinfo')
        continuing = self.data.get('continue')

        if missing and not deferred and not continuing:
            self.get_imageinfo(show=False)

    def _update_params(self):
        """
        update params from response data
        """
        if self.data.get('title'):
            self.params['title'] = self.data.get('title')
        if self.data.get('pageid'):
            self.params['pageid'] = self.data.get('pageid')
        if self.data.get('wikibase'):
            self.params['wikibase'] = self.data.get('wikibase')

    def skip_action(self, action):
        """
        append action to skip flag
        """
        if 'skip' not in self.flags:
            self.flags['skip'] = []
        self.flags['skip'].append(action)

    def get(self, show=True, proxy=None, timeout=0):
        """
        Make Mediawiki, RESTBase, and Wikidata requests for page data
        some sequence of:
        - get_parse()
        - get_query()
        - get_restbase()
        - get_wikidata()
        """
        wikibase = self.params.get('wikibase')

        if wikibase:

            self.flags['defer_imageinfo'] = True

            self.get_wikidata(False, proxy, timeout)
            self.get_query(False, proxy, timeout)
            self.get_parse(False, proxy, timeout)

            self.flags['defer_imageinfo'] = False

            self.get_restbase('/page/summary/', False, proxy, timeout)

            if show and not self.flags.get('silent'):
                self.show()

        else:

            self.flags['defer_imageinfo'] = True

            self.get_query(False, proxy, timeout)
            self.get_parse(False, proxy, timeout)

            if not self.data.get('wikibase'):
                self.skip_action('wikidata')

            self.get_wikidata(False, proxy, timeout)

            self.flags['defer_imageinfo'] = False

            wiki = self.params.get('wiki')
            if wiki and 'wikipedia.org' not in wiki:
                self.skip_action('restbase')

            self.get_restbase('/page/summary/', False, proxy, timeout)

            if show and not self.flags.get('silent'):
                self.show()

        return self

    def get_imageinfo(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki request for API:Imageinfo
        https://www.mediawiki.org/wiki/API:Imageinfo

        Required {data}:
        - image: <list> member (<dict>) with 'file' and not 'url'

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - image: <list> member (<dict>) image URLs, sizes, etc.
        """
        if not self.data.get('image'):
            raise ValueError("get_imageinfo needs a page image")

        if not self._missing_imageinfo() and 'imageinfo' in self.cache:
            utils.stderr("complete imageinfo in cache", self.flags['silent'])
            return

        self._normalize_images()
        self._get('imageinfo', show, proxy, timeout)

        return self

    def get_more(self, show=True, proxy=None, timeout=0):
        """
        Calls get_querymore() Is for convenience. You like.
        """
        return self.get_querymore(show, proxy, timeout)

    def get_parse(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki:API action=parse request
        https://en.wikipedia.org/w/api.php?action=help&modules=parse

        Required {params}: title OR pageid
        - title: <str> article title
        - pageid: <int> Wikipedia database ID

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - image: <dict> {parse-image, parse-cover}
        - infobox: <dict> Infobox data as python dictionary
        - iwlinks: <list> interwiki links
        - pageid: <int> Wikipedia database ID
        - parsetree: <str> XML parse tree
        - requests: list of request actions made
        - wikibase: <str> Wikidata entity ID or wikidata URL
        - wikitext: <str> raw wikitext URL
        """
        if not self.params.get('title') and not self.params.get('pageid'):
            raise ValueError("get_parse needs title or pageid")

        self._get('parse', show, proxy, timeout)

        return self

    def get_query(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki:API action=query selected data
        https://en.wikipedia.org/w/api.php?action=help&modules=query

        Required {params}: title OR pageid
        - title: <str> article title
        - pageid: <int> Wikipedia database ID

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - description: <str> Wikidata description (via pageterms)
        - extext: <str> plain text (Markdown) extract
        - extract: <str> HTML extract from Extension:TextExtract
        - image: <dict> {query-pageimage, query-thumbnail}
        - label: <str> Wikidata label (via pageterms)
        - modified (page): <str> ISO8601 date and time
        - pageid: <int> Wikipedia database ID
        - random: <str> a random article title with every request!
        - requests: list of request actions made
        - url: <str> the canonical wiki URL
        - url_raw: <str> ostensible raw wikitext URL
        - watchers: <int> number of people watching this page
        """
        if not self.params.get('title') and not self.params.get('pageid'):
            raise ValueError("get_query needs title or pageid")

        self._get('query', show, proxy, timeout)

        while self.data.get('continue'):
            self._get('query', show, proxy, timeout)

        return self

    def get_querymore(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki:API action=query for MORE data
        A much more expensive (slower!) query for popular pages
        https://en.wikipedia.org/w/api.php?action=help&modules=query

        Required {params}: title OR pageid
        - title: <str> article title
        - pageid: <int> Wikipedia database ID

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - categories: <list> list of categories used on the page
        - contributors: <int> total number of contributors
        - files: <list> list of files contained in the page
        - languages: <int> number of wiki languages having this page
        - views: <int> average daily page views over past 60 days
        """
        if not self.params['title'] and not self.params['pageid']:
            raise ValueError("get_query needs title or pageid")

        self._get('querymore', show, proxy, timeout)

        while self.data.get('continue'):
            self._get('querymore', show, proxy, timeout)

        return self

    def get_random(self, show=True, proxy=None, timeout=0):
        """
        GET MediaWiki:API (action=query) list=random
        https://www.mediawiki.org/wiki/API:Random

        Required {params}: None

        Optional arguments:
        - [show]: <bool> echo page data if true
        - [proxy]: <str> use this HTTP proxy
        - [timeout]: <int> timeout in seconds (0=wait forever)

        Data captured:
        - pageid: <int> Wikipedia database ID
        - title: <str> article title
        """
        self._get('random', show, proxy, timeout)

        # flush cache to allow repeated random requests
        del self.cache['random']

        return self

    def images(self, fields=None, token=None):
        """
        Returns image info keys for kind containing token

        Args:
        - fields: <list> image info values wanted
        - token: <str> substring to match image kind

        EXAMPLES

        Get all available image info:
        >>> page.images()

        Get all image kinds:
        >>> page.images('kind')

        Get only "query" (kind) image info
        >>> page.images(token='query')

        Get only file, url fields from "wikidata" images
        >>> page.images(['file', 'url'], token='wikidata')
        """

        if 'image' not in self.data:
            return

        out = []
        for img in self.data['image']:
            if token and token not in img['kind']:
                continue
            info = {}
            for key in img:
                if fields and key not in fields:
                    continue
                info.update({key: img[key]})
            if info:
                out.append(info)

        return out

    def pageimage(self, token=None):
        """
        DEPRECATED see page.images()
        """
        return self.images(token=token)
